package ipn;

import java.util.Collection;
import java.util.HashSet;
import java.util.Vector;

import libpetri.nets.IPTNet;

/** An IPN model or scenario.
 * 
 * Contains a set of type declarations, a main instance and an initial state for the main instance.
 * Operations to build state space, convert to simple PT net, and export to IPN file format provided.
 * 
 * @author yann
 *
 */
public class IPNModel {

	/** The types of this model */
	private Collection<IPNType> types= new Vector<IPNType>();
	
	/** The main instance */
	private IPNInstance main;
	
	/** The initial state */
	private String initialState = "default";
	
	/** Define the main instance.
	 * 
	 * @param typeName : the type of the main instance.
	 * @param instanceName : a name for the main instance.
	 * @return false + stderr message if the type does not exist.
	 */
	public boolean setMain(String typeName, String instanceName) {
		IPNType t = findType(typeName); 
		if (t == null) {
			System.err.println("Unknown type "+typeName+ " when handling model.setMain("+typeName+","+instanceName+")");
			return false;
		}
		main = new IPNInstance(instanceName,t,0);
		return true;
	}
	
	/** Set the initial state of the main instance.
	 * 
	 * @param stateName : the name of the state, will default to "default" initial state.
	 * @return true, as no test to ensure existence of state is done.
	 */
	public boolean setInitialState (String stateName) {
		initialState = stateName;
		return true;
	}

	/** Add a type declaration, provided a type of that name does not preexist.
	 * 
	 * @param type : reference to the new type.
	 * @return true if success, false + stderr message else.
	 */
	public boolean addType (IPNType type) {
		if (findType(type.getName()) != null) {
			System.err.println("Type "+type.getName()+" already exists in model when handling addType.");
			return false;
		}
		types.add(type);
		return true;
	}

	/** Find a type declaration based on its name.
	 * 
	 * @param typeName : name of the type to find.
	 * @return null if type does not exist.
	 */
	public IPNType findType (String typeName) {
		for (IPNType t : types) {
			if (t.getName().equals(typeName))
				return t;
		}		
		return null;
	}
	
	/** Convert to a flat PT net representation.
	 * 
	 * @param receiver : the net receiver will be updated to contain a model equivalent to this one.
	 */
	public void flatten (IPTNet receiver) {
		main.getType().flatten(receiver,main.getName(), main.getType().getState(initialState));
	}
	
	/** Construct the state space of the IPN, using basic explicit reachability algorithm.
	 * 
	 * Prints some statistics + deadlock states to stdout.
	 */
	public void buildStateSpace () {	
		HashSet<IPNState> seen = new HashSet<IPNState>();
		Vector<IPNState> todo = new Vector<IPNState>();
		IPNState init = main.getType().getState(getInitialState());
		//System.out.println(init.toString(this));		
		seen.add(init);
		todo.add(init);
		while ( ! todo.isEmpty()) {
			IPNState s = todo.remove(0);
//			System.out.println("next state :" + s);
			Collection<IPNState> succ = main.getType().getPrivateSuccessors(s.copy());
			for (IPNState ns : succ) {
				if (seen.add(ns)) {
//					System.out.println( ns.toString());
					todo.add(ns);
				}
			}
			if (succ.isEmpty()) {
				System.out.println("Found a deadlock state ! :");
			}
		}

		System.out.println("Seen " + seen.size() + "\n");
//		for (State s : seen) {
//		System.out.println(s.toString(this)+"\n-----------------------------------\n");
//		}
		
	}
	
	/** Accessor for initial state of main instance.
	 * 
	 * @return the name of the initial state (or "default" if unset).
	 */
	public String getInitialState() {
		return initialState;
	}
	
	/** Return a string representation of this model, in IPN file format style.
	 * 
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		for (IPNType t : types) {
			sb.append(t.toString());
		}
		if (main != null) {
			sb.append("// the main model instance");
			sb.append("\nmain "+main.getType().getName()+ " "+main.getName() + ";\n");
		}
		return sb.toString();
	}

	public IPNInstance getMain() {
		return main;
	}
}
