package ipn.elementary;

import java.util.Arrays;

import libpetri.structure.state.UniqueArrayHelper;

import ipn.IPNState;
import ipn.IPNType;




/** A basic implem of a net state.
 * 
 * @author yann
 * 
 * Contains a beautiful hidden mechanism to ensure state uniqueness at any depth.
 * Protocol : get/copy/set* /unique ;
 *  * NetState.copy ->Helper and Helper.unique ->NetState
 * set on Uniquehelper handles dirty flags and late copy of data.
 * unique relies on a static table ! BE WARNED THIS IS NOT thread-safe
 * if you want thread safe basic implem of states, look at revision 1610. 
 * or just set unique flag to false
 */
public class Marking implements IPNState, Cloneable {
	/// to store place states
	protected int marks[];

	/// initialize sizes w.r.t. n
	public Marking (UnitaryType n) {
		marks = new int[n.places.size()];
	}
	/// for use by UniqueHelper
	protected Marking(int marks[]) {
		this.marks = marks;
	}
	/// public copy semantics constructor
	public Marking(Marking state) {
		marks = state.marks.clone();			
	}
	
	public void setPlaceState(Place p,int value) {	
		setPlaceState(p.getIndex(), value);
	}
	/// modify place state. Size checks are made.
	protected void setPlaceState(int index,int value) {		
		ensureCapacity(index+1);
		marks[index]=value;
	}

	///  function to ensure capacity, inspired by java.util.Vector code
	private void ensureCapacity (int minCapacity) {
		int oldCapacity = marks.length;
		if (minCapacity > oldCapacity) {
			int[] oldData = marks;
			marks = new int[minCapacity];
			System.arraycopy(oldData, 0, marks, 0, oldData.length);
		}
	}

	public int getPlaceState(Place place) {
		int index = place.getIndex();
		return getPlaceState(index);
	}

	public Marking copy() {
		return new UniqueHelper(this);
	}

	private int getPlaceState(int index) {
		ensureCapacity(index+1);
		return marks[index];
	}

	@Override
	public String toString() {
		return Arrays.toString(marks);
	}
	
	public String toString (IPNType n) {
		UnitaryType net = (UnitaryType) n;
		StringBuffer sb = new StringBuffer();
		for (Place p:net.places) {
			int m = getPlaceState(p);
			if (m != 0)
				sb.append("\t  "+p.getName()+" = "+m +";\n");
		}
		return sb.toString();
	}


	@Override
	public int hashCode() {
		return Arrays.hashCode(marks);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (! (obj instanceof Marking) )
			return false;
		Marking other = (Marking) obj;
		if (!Arrays.equals(marks, other.marks))
			return false;
		return true;
	}

	public Marking unique(IPNType net) {
		UniqueHelper uh = new UniqueHelper(this);
		return uh.unique(net);
	}

	public static String getUnicityStats () {
		return UniqueHelper.toStringUniqueTable();
	}
	public int compareTo(IPNState s2) {		
		int ret=0;
		if (s2 instanceof Marking) {
			Marking s = (Marking) s2;

			// check place vector size
			ret = marks.length - s.marks.length;
			if (ret != 0) {
				return ret/ Math.abs(ret);
			}		
			// lexico on marks
			for (int i=0;i < marks.length;i++) {
				ret = marks[i] - s.marks[i];
				if (ret !=0)
					return ret/ Math.abs(ret);			
			}
		} else {
			return 1;
		}
		return ret;
	}
}

class UniqueHelper extends Marking {

	// true if place marks has been touched
	private boolean isDirtyMarks;

	public UniqueHelper(Marking state) {
		// NO COPY using protected constructor
		super(state.marks);
		// everyone is clean
		isDirtyMarks = false;
	}

	public Marking unique(IPNType net) {
		if (isDirtyMarks)
			return new Marking(UniqueArrayHelper.unique(marks));
		else
			return new Marking(marks);
	}


	@Override
	// set the dirty flags and copy the data just before modifying
	public void setPlaceState(int index, int value) {
		if (!isDirtyMarks) {
			isDirtyMarks = true;
			marks = marks.clone(); 
		}
		super.setPlaceState(index, value);
	}

	// for debug, pretty print of unique tables
	public static String toStringUniqueTable() {
		StringBuffer sb = new StringBuffer();
		sb.append("Markings Unicity table.\n");
		sb.append(UniqueArrayHelper.getStats());
		return sb.toString();		
	}

}



