package libpetri.nets;

import java.util.Iterator;

/** An interface materializing the notion of Labeled transition system.
 * 
 * An LTS is a tuple <S,L,Enab,Tau> where 
 * * S is the set of states of the LTS.
 * * L = L_pub \cup L_priv is the set of labels of the LTS, partitioned into 
 * ** L_pub publicly visible labels (interface of the LTS)
 * ** L_priv private labels often noted \tau or \epsilon not used for synchronization.
 * * Enab \subset SxL : is the enabling function such that l \in L is said to be enabled in state s \in S iff. <s,l> \in Enab.   
 * * Tau : Enab -> S is the partial deterministic transition function which maps a <state,label> to a successor state. 
 * @author thierry
 *
 * @param <State> : the state kind.
 */
public interface ILabeledTransitionSystem<State> {
	
	/** Accessor allowing to traverse the full set of labels of this LTS */
	public Iterator<String> getLabels ();
	/** Accessor allowing to traverse the set of public labels of this LTS */
	public Iterator<String> getPublicLabels ();
	/** Accessor allowing to traverse the set of private labels of this LTS */
	public Iterator<String> getPrivateLabels ();
	
	/** Allows to check existence of a given label */
	public boolean containsLabel (String label);
	/** Returns true if the label is enabled in state source */
	public boolean isEnabled (String label, State source);
	/** Returns the successor state of this LTS by firing of the event labeled by label. 
	 *  Note that the result is not well defined if the label is not enabled in the source state.*/
	public State getSuccessor (String label, State source);	

}
