package libpetri.structure;

import java.util.Collection;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Vector;

import libpetri.graphics.AnnotationGraphics;
import libpetri.graphics.ArcGraphics;
import libpetri.graphics.NodeGraphics;
import libpetri.namespaces.NamedElement;
import libpetri.namespaces.Namespace;
import libpetri.namespaces.NamingException;
import libpetri.nets.ICompositeNet;
import libpetri.nets.ISymmetricNet;
import libpetri.structure.arcs.Arc;
import libpetri.structure.arcs.ArcType;
import libpetri.structure.arcs.Function;
import libpetri.structure.arcs.ParseFunction;
import libpetri.structure.state.NetState;
import libpetri.structure.state.ParseMarking;
import libpetri.structure.state.PlaceState;
import libpetri.structure.state.State;

public class Net extends Namespace implements ISymmetricNet,ICompositeNet<Net> {

	Vector<Place> places;
	Vector<Transition> transitions;
	State initialState;
	Vector<Net> components;
	Vector<SynchTransition> sync;
	Vector<TypeDef> types;
	Vector<Domain> doms;

	public Net(String name) {
		super(name);
		places = new Vector<Place>();
		transitions = new Vector<Transition>();
		initialState = new NetState(this);
		sync = new Vector<SynchTransition>();
		components = new Vector<Net>();
		types = new Vector<TypeDef>();
		doms = new Vector<Domain>();
		doms.add(Domain.getNullDomain());
	}


	public boolean addSynchronize(String tName, String slaveName) {
		return addSynchronize(tName, slaveName, "()");
	}

	public boolean addSynchronize(String tName, String slaveName, String param) {
		try {
			ITransition slave = (ITransition) find(slaveName);
			SynchTransition t = (SynchTransition) find(tName);
			if (slave != null && t!=null) {
				return t.addSynchronization(slave,param);				
			} else {
				System.err.println("Object not found : tName:"+t+  " slave:"+slave);
				return false;
			}
		} catch (ClassCastException e) {
			e.printStackTrace();
			return false;
		}
	}

	public boolean createArc (String pname, String tname, ArcType at) {
		return createArc(pname,tname,at,1);
	}

	public boolean createArc(String pname, String tname, ArcType at, int value) {
		return createArc(pname, tname, at,value+"'<>");
	}
	
	public boolean createArc(String pname, String tname, ArcType at, String fun, ArcGraphics g) {
		Place p = null;
		Transition t= null;
		try {
			p = (Place) find(pname);
			if (p == null) {
				System.err.println("Unknown place "+pname+ " when treating createArc("+pname+","
						+tname+","+at+","+fun+")");
				return false;
			}
			t = (Transition) find(tname);
			if (t == null) {
				System.err.println("Unknown transition "+tname+ " when treating createArc("+pname+","
						+tname+","+at+","+fun+")");
				return false;
			}
			Function f = ParseFunction.parse(p,t,fun);
			Arc a = Arc.createArc(p, t, at,f);
			a.setGraphics(g);
			t.addArc(a);
			p.addArc(a);
			return true;
		} catch (ClassCastException e) {
			System.err.println ("Failed to create arc linking place : "+pname 
					+ " to trans :"+tname + "\n in fact p="+p + " t="+ t);
		}
		return false;	
	}


	public boolean createArc(String pname, String tname, ArcType at, int value, ArcGraphics g) {
		return createArc(pname, tname, at,value+"'<>",g);
	}

	public boolean createArc(String pname, String tname, ArcType at, String fun) {
		return createArc(pname, tname, at, fun, null);
	}

	public boolean createComponent(Net n) {
		try {
			add(n);
			initialState.setComponentState(components.size(), n.getInitialState());
			return components.add(n);
		} catch (NamingException e) {
			e.printStackTrace();
			return false;
		}
	}


	public boolean createTypeDef(String typename, String clinames) {
		try {
			TypeDef t = new TypeDef(typename,clinames);
			add(t);
			types.add(t);
			return true;
		} catch (NamingException e) {
			e.printStackTrace();
			return false;
		}
	}

	public boolean createDomain(String domName, String domElts) {
		try {
			Domain d = new Domain(domName);
			for (String name : domElts.split(",")) {
				TypeDef type = (TypeDef)find(name);
				if (type == null) {
					System.err.println("Name "+name+" does not correspond to a basic type when treating createDomain("+domName+","+domElts+")");
					return false;
				}
				d.add(type);
			}
			doms.add(d);
			add(d);
			return true;
		} catch (ClassCastException e) {
			System.err.println("Name does not correspond to a basic type when constructing domain "+domName);
			e.printStackTrace();
			return false;
		} catch (NamingException e) {
			e.printStackTrace();
			return false;
		}
	}


	public boolean createPlace(String name) {				
		Place p= new Place(name);			
		return addPlace(p);
	}

	private boolean addPlace (Place p) {
		try {				
			super.add(p);
			places.add(p);
			p.setIndex(places.size()-1);
			//ensure capcity
			setPlaceState(p.getName(), "0");
			return true;
		} catch (NamingException e) {
			System.err.println("Naming error !"+e);
			e.printStackTrace();
			return false;
		}
	}

	/** Creates a place of this domain.
	 * 
	 * @param place : the place name
	 * @param domain : the place domain
	 */
	public boolean createPlace(String place, String domain) {
		try {
			Domain dom = (Domain)find(domain);
			if (dom == null) {
				System.err.println("Unknown domain "+domain+ " when treating createPlace("+place+","+domain+")");
				return false;
			}
			Place p = new Place(place,dom);

			return addPlace(p);
		} catch (ClassCastException e) {
			System.err.println("Cannot create a place of domain "+domain+ " as this name corresponds to "+find(domain));
			e.printStackTrace();
			return false;
		}
	}


	public boolean createSynchronization(String name, Visibility vis) {
		SynchTransition t= new SynchTransition(name,vis);		
		try {
			super.add(t);
			sync.add(t);
		} catch (NamingException e) {
			e.printStackTrace();
			return false;
		}
		return true;
	}

	/* (non-Javadoc)
	 * @see structure.IPTNet#createTransition(java.lang.String, structure.Visibility)
	 */
	public boolean createTransition(String name, Visibility vis) {
		try {
			Transition t= new Transition(name,vis);			
			super.add(t);
			transitions.add(t);
			t.setIndex(transitions.size()-1);
			return true;
		} catch (NamingException e) {
			System.err.println("Naming error !"+e);
			e.printStackTrace();
			return false;
		}
	}

	public boolean createTransitionParameter(String tname,String varName,String type) {
		try {
			Transition t = (Transition) find(tname);
			if (t == null) {
				System.err.println("Unknown transition "+tname+ " when treating createTransitionParameter("+
						tname+","+varName+","+type+")");
				return false;
			}
			TypeDef typ = (TypeDef)find(type);
			if (typ == null) {
				System.err.println("Unknown type name "+type+ " when treating createTransitionParameter("+
						tname+","+varName+","+type+")");
				return false;
			}
			t.addParameter(new Variable(varName,typ));
		} catch (Exception e) {
			System.err.println("Error while creating transition parameter " + e);
			e.printStackTrace();
		}
		return true;
	}

	public void generateStateSpace() {
		HashSet<State> seen = new HashSet<State>();
		Vector<State> todo = new Vector<State>();
//		PlaceState ps = ParseMarking.parse(places.get(0),"1");
//		System.out.println(ps.toString(places.get(0).getDom()));		
//		State init = new NetState(this);
//		init.addToPlaceState(0, ps);
		State init = getInitialState();
		System.out.println(init.toString(this));		
		seen.add(init);
		todo.add(init);
		while ( ! todo.isEmpty()) {
			State s = todo.remove(0);
			Enumeration<State> succ = getSuccessors(s);
			int nbSucc = 0;
			while (succ.hasMoreElements()) {
				State ns = succ.nextElement();
				if (seen.add(ns)) {
//					System.out.println(ns + "   " + ns.toString(this));
					todo.add(ns.copy());
				}
				nbSucc++;
			}
			if (nbSucc == 0) {
				System.out.println("Found a deadlock state ! :" + s.toString(this));
			}
		}

		System.out.println("Seen " + seen.size() + "\n");
//		for (State s : seen) {
//		System.out.println(s.toString(this)+"\n-----------------------------------\n");
//		}

	}


	protected String getCompDecl() {
		if (components.isEmpty())
			return "";
		StringBuffer sb = new StringBuffer();
		sb.append("\n\t// Components");
		for (Enumeration<NamedElement> e = elements(); e.hasMoreElements() ; ) {
			NamedElement ne = e.nextElement();
			if (ne instanceof Net) {
				Net n = (Net) ne;
				sb.append("\n\tNet "+n.getName() + " : " + n.getClass().getSimpleName() + ";");
			}
		}
		return sb.toString();
	}

	public int getComponentIndex(String name) {
		for (int i=0; i< components.size(); i++) {
			if (components.get(i).name.equals(name)) {
				return i;
			}
		}
		return -1;
	}

	public Collection<Net> getComponents() {
		return components;
	}

	protected String getDecl() {
		return getTypeDecl() +getCompDecl() + getPlaceDecl() + getTransDecl() + getSyncDecl();
	}


	private String getTypeDecl() {
		StringBuffer sb = new StringBuffer();
		if (! types.isEmpty()) {
			sb.append("\n\t //  Type definitions ");
			for (TypeDef t : types) {
				sb.append("\n\t"+t.toString());
			}
			sb.append("\n\n\t //  Domain definitions ");
			// skip domain 0 = null domain
			for (int i =1 ; i < doms.size(); i++) {
				sb.append("\n\t"+doms.get(i).toString());
			}
		}
		return sb.toString();
	}

	public String getFooter () {
		return "\n}";
	}

	public String getHeader () {
		return ("\nNet "+name +":"+ getClass()+ "{\n");
	}

	public final State getInitialState() {
		initialState = initialState.unique(this);
		return initialState;
	}

	public String getPlaceDecl () {
		if (places.isEmpty()) 
			return "";
		StringBuffer sb = new StringBuffer();
		sb.append("\n\n\t// Places"); 
		for (Place p : places) {
			sb.append(p);
			PlaceState m = initialState.getPlaceState(p);
			if (! m.isEmpty()) {
				sb.append("\n\t"+p.getName()+" = ");
				sb.append(m.toString(p.getDom()));
			}
		}
		return sb.toString();
	}

	public Collection<State> getPrivateSuccessors(State src) {
		Vector<State> succ= new Vector<State>();
		for (ITransition t:transitions) {
			if (t.getVis()==Visibility.PRIVATE) {
				BindingSet bs = t.getPotentialBindings();
				bs = t.isEnabled(src,bs);
				Enumeration<State> sit = t.getSuccessors(src, bs);
				while (sit.hasMoreElements()) {
					State s2 = sit.nextElement();
					succ.add(s2);
				}
			}
		}
		return succ;
	}

	public Enumeration<State> getSuccessors(State src) {
		Vector<State> succ = new Vector<State>();
		for (ITransition t:sync) {
			if (t.getVis()==Visibility.PRIVATE) {
				BindingSet bs = t.getPotentialBindings();
				bs = t.isEnabled(src,bs);
				Enumeration<State> sit = t.getSuccessors(src, bs);
				while (sit.hasMoreElements()) {
					State s2 = sit.nextElement();
					succ.add(s2);
				}
			}
		}
		int i =0;
		for (Net n:components) {
			Enumeration<State> localsucc = (n.getSuccessors(src.getComponentState(i)));
			while ( localsucc.hasMoreElements() ) {
				State loc = localsucc.nextElement();
				State s2 = src.copy();
				s2.setComponentState(i, loc);
				succ.add(s2.unique(this));
			}
			i++;
		}
		succ.addAll(getPrivateSuccessors(src));
		return succ.elements();
	}

	protected String getSyncDecl() {
		if (sync.isEmpty())
			return "";
		StringBuffer sb = new StringBuffer();
		sb.append("\n\n\t// Synchronization Transitions");
		for (SynchTransition t: sync) {
			sb.append(t);
		}
		return sb.toString();
	}

	private String getTransDecl() {
		if (transitions.isEmpty())
			return "";
		StringBuffer sb = new StringBuffer();
		sb.append("\n\t// Transitions");
		for (ITransition t:transitions) {
			sb.append(t);
		}
		return sb.toString();
	}


	public boolean setPlaceState(String name, String mark) {
		Place p = (Place) find(name);		
		if (p==null) {
			System.err.println("Error while treating setPlaceState("+name+","+mark+") : no place named "+name+" found");
			return false;
		}
		PlaceState f = ParseMarking.parse(p,mark);
		//	System.out.println(f.toString(p.getDom()));
		initialState.addToPlaceState(p,f);
		return true;
	}

	public String toString() {
		return getHeader() + getDecl () + getFooter() ;		
	}



	public void flatten (Net receiver) {
		for (Net n:components) {
			n.flatten(receiver);
		}
		for (Place p : places) {
			String n = p.getFullyQualified_Name();
			receiver.createPlace(n);
			receiver.setPlaceState(n, initialState.getPlaceState(p).toString(p.getDom()));
		}
		for (Transition t : transitions) {
			if (t.getVis() == Visibility.PRIVATE) {
				t.flatten(receiver,null);
			}
		}
		for (SynchTransition t : sync) {
			if (t.getVis() == Visibility.PRIVATE) {
				t.flatten(receiver,null);
			}
		}
	}

	public void convert (ISymmetricNet a) {
		if (! components.isEmpty()) {
			System.err.println("WARNING : Export to AMI net format is currently unsupported " +
			"for hierarchical definitions. Use unfold option before export. Continuing with loss of information.");
		}
		for (Place p:places) {
			a.createPlace(p.getName());
			a.setPlaceState(p.getName(), initialState.getPlaceState(p).toString(p.getDom()));
		}
		for (Transition t : transitions) {
			t.convert(a);
		}
	}


	public Collection<ITransition> getPublicTransitions() {
		Vector<ITransition> v = new Vector<ITransition>();
		for (ITransition t : transitions) {
			if (t.getVis() == Visibility.PUBLIC)
				v.add(t);
		}
		for (ITransition t : sync) {
			if (t.getVis() == Visibility.PUBLIC)
				v.add(t);
		}
		return v;
	}

	public boolean setPlaceState(String name, int mark) {
		return setPlaceState(name,Integer.toString(mark));
	}


	public Collection<Place> getPlaces() {
		return places;
	}





	


	public NodeGraphics getNodeGraphics(String nodeName) {
		try {
			Node n = (Node) find(nodeName);
			if (n == null) {
				System.err.println("Unknown node "+nodeName+ " when treating getNodeGraphics");
				return null;
			} else
				return n.getGraphics();
		} catch (ClassCastException e) {
			System.err.println("Object  "+nodeName+ " is not a node when treating getNodeGraphics. ");
			return null;
		}
	}


	public void setNodeAnnotationGraphics(String nodeName, String attribName, AnnotationGraphics g) {
		try {
			Node n = (Node) find(nodeName);
			if (n == null) {
				System.err.println("Unknown node "+nodeName+ " when treating getNodeGraphics");
				return ;
			} else
				n.setTagPosition(attribName, g);
		} catch (ClassCastException e) {
			System.err.println("Object  "+nodeName+ " is not a node when treating setNodeGraphics. ");
			return ;
		}
	}
	
	public AnnotationGraphics getNodeAnnotationGraphics(String nodeName, String attribName, AnnotationGraphics g) {
		try {
			Node n = (Node) find(nodeName);
			if (n == null) {
				System.err.println("Unknown node "+nodeName+ " when treating getNodeGraphics");
				return null;
			} else
				return n.getTagPosition(attribName);
		} catch (ClassCastException e) {
			System.err.println("Object  "+nodeName+ " is not a node when treating setNodeGraphics. ");
			return null ;
		}
	}


	public void setNodeGraphics(String nodeName, NodeGraphics g) {
		try {
			Node n = (Node) find(nodeName);
			if (n == null) {
				System.err.println("Unknown node "+nodeName+ " when treating setNodeGraphics");
				return ;
			} else
				n.setGraphics(g);
		} catch (ClassCastException e) {
			System.err.println("Object  "+nodeName+ " is not a node when treating setNodeGraphics. ");
			return ;
		}
	}	



}
