package libpetri.structure;

import java.util.Enumeration;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import libpetri.namespaces.NamedElement;
import libpetri.nets.ISymmetricNet;
import libpetri.structure.state.State;

public class SynchTransition extends NamedElement implements ITransition {
	
/**
 * @label contains thru synch
 */
/*#structure.BoundTransition Dependency20*/


	Vector<ITransition> synch;
	private Visibility vis;
	private Vector<Variable> param;

	public SynchTransition(String name, Visibility vis) {
		super(name);
		synch = new Vector<ITransition>();
		param = new Vector<Variable>();
		this.vis = vis;
	}

	public BindingSet getPotentialBindings() {
		// TODO no variables available yet
		return new BindingSet(new Vector<Variable>());
	}


	public Enumeration<State> getSuccessors(State s, BindingSet bs) {
		return new SuccIterator(this,bs,s);
	}

	public void fire(State s, int[] bind) {
		Net n = (Net) getNamespace();
		for (ITransition t : synch) {
			int i = n.getComponentIndex(t.getContainer().getName());					
			State s2 = s.getComponentState(i).copy();
			t.fire(s2,bind);
			s.setComponentState(i, s2.unique(n));
		}
	}

	public BindingSet isEnabled(State s, BindingSet bs) {
		if (bs.isEmpty())
			return bs;
		Net n = getContainer();
		for (ITransition t : synch) {
			int i = n.getComponentIndex(t.getContainer().getName());					
			bs = t.isEnabled(s.getComponentState(i),bs);
			// break if not enabled anymore
			if (bs.isEmpty()) 
				return bs;
		}
		return bs;
	}


	public boolean addSynchronization (ITransition t, String param) {

		int [] adapt = null;
		Pattern pat = Pattern.compile("^\\s*\\(\\s*(.*?)\\s*\\)\\s*$");
		Matcher matcher = pat.matcher(param);
		if (matcher.find()) {
			if (matcher.group(1) != null && ! matcher.group(1).equals("")) {
				String [] vars = matcher.group(1).split("\\s*,\\s*");
				adapt = new int[vars.length];
				int i = 0;
				for (String var : vars) {
					adapt[i++] = getVarIndex(var);
				}
			} else {
				adapt = new int[0];
			}
		} else {
			System.err.println("Error parsing parameter string in "+this.getName()
					+" addSynchronization(transition ="+t.getName()+", param ="+param+")");
			return false;
		}

		return synch.add(new BoundTransition(adapt,t));

	}

	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("\n\t");
		switch (vis) {
		case PRIVATE : sb.append("private "); break;
		case PUBLIC : sb.append("public ");
		}
		sb.append("synchronization "+getName()+ " {");
		for (ITransition t : synch)
		//	sb.append("\n\t\t synchronize "+((NamedElement)t).getFullyQualifiedName().split("\\.",2)[1]+" ;");
			sb.append("\n\t\t synchronize "+ t.getContainer().getName() + "." + t.getName()+" ;");

		sb.append("\n\t}");
		return sb.toString();
	}


	public Visibility getVis() {
		return vis;
	}


	public void flatten(ISymmetricNet receiver,String alias) {
		if (alias==null) {
			alias = getFullyQualified_Name();
			receiver.createTransition(alias, Visibility.PRIVATE);
		}
		for (ITransition t:synch) {			
			t.flatten(receiver,alias);
		}
	}


	public Net getContainer() {
		return (Net) getNamespace();
	}

	public void addParameter(Variable variable) {
		param.add(variable);
		variable.setIndex(-param.size());
	}

	public int getVarIndex (String name) {
		for (Variable v : param) {
			if (v.getName().equals(name))
				return v.getIndex();
		}
		return 0;
	}

	public Variable getVarByIndex(int id) {
		for (Variable v : param) {
			if (v.getIndex() == id )
				return v;
		}
		return null;
	}

}

class BoundTransition implements ITransition {
	int [] adapter;
	int [] effective;
	ITransition  t;
	public BoundTransition(int[] adapt, ITransition t2) {
		adapter = adapt;
		effective = new int[adapt.length];
		t = t2;
	}

	public void fire(State s, int[] bind) {
		for (int i = 0 ; i < adapter.length ; i++ ) {
			effective[i] = bind[adapter[i]];
		}
		t.fire(s, effective);
	}

	public void flatten(ISymmetricNet receiver, String alias) {
		t.flatten(receiver, alias);
	}
	public Net getContainer() {
		return t.getContainer();
	}
	public String getName() {
		return t.getName();
	}
	public BindingSet getPotentialBindings() {
		return t.getPotentialBindings();
	}
	public Enumeration<State> getSuccessors(State s, BindingSet bs) {
		bs.setAdapter(adapter);
		return t.getSuccessors(s, bs);
	}
	public Visibility getVis() {
		return t.getVis();
	}

	public BindingSet isEnabled(State s, BindingSet bs) {
		bs.setAdapter(adapter);
		return t.isEnabled(s, bs);
	}

}
