package libpetri.tensor;

import java.util.Arrays;
import java.util.BitSet;

/** Un objet mathmatique multidimensionnel. 
 * Un tenseur de rang 1 = vecteur, rang 2 = matrice, rang 3 = cube , rang 4 = hypercube....
 * La reprsentation s'appuie sur un bitset, donc ca dechire en perfs et memoire.
 * !!!!!!!!!! Actuellement limit a valeurs dans 0/1. 
 * Evolution vers EnumSet de 0,1,2,PLUS (i.e. sur 2 bits) et stockage
 *  en tableaux annexes des valeurs > 2 envisage.
 */
public class Tensor implements Cloneable {

	/** The dimension of each rank. dims.length = rank of the tensor.*/
	int [] dims;
	/** to store the tensor. Currently only supports 0/1 values ("color safe" nets)*/
	BitSet vals;

	/** construct an empty tensor given its rank and dimensions. 
	 * !! dims is not copied, but referenced from the tensor. 
	 * Thus dims should not be modified after tensor construction.
	 * dims = [] constructs a null rank tensor that cannot hold any elements.
	 * dims = [4] constructs a vector of size 4.
	 * dims = [3,2] constructs a 3x2 matrix.  
	 * etc...
	 * */
	public Tensor(int [] dims) {
		this.dims = dims;
		int s=1;
		for (int i:dims)
			s*=i;
		vals = new BitSet(s);
	}

	/** Internal constructor, no checks performed */
	Tensor(int[] dims2, BitSet vals) {
		this.dims = dims2;
		this.vals = vals;
	}

	/** tests whether the tensor contains the element provided.
	 * 
	 * @param coords gives a definition of the element, 
	 * i.e. the coordinates of a point in the tensor space along each axis of the tensor. 
	 * @return true if the element is already in the tensor.
	 */
	public boolean contains(int [] coords ){		
		return vals.get(toAdr(coords));
	}
	
	/** Adds the element provided to a tensor.
	 * 
	 * @param coords gives a definition of the element, 
	 * i.e. the coordinates of a point in the tensor space along each axis of the tensor. 
	 */
	public void add (int [] elt) {
		vals.set(toAdr(elt));
	}
	
	/** Removes an occurrence of the element provided from a tensor.
	 * 
	 * @param coords gives a definition of the element, 
	 * i.e. the coordinates of a point in the tensor space along each axis of the tensor. 
	 */
	public void remove(int[] elt) {
		vals.clear(toAdr(elt));		
	}

	/** Used internally to translate coordinates into an index in the bitset. 
	 *  See also toCoord() for reverse operation.
	 *  Uses this.dims for adress computations.
	 * */
	private int toAdr (int [] coords) {
		return toAdr(coords,dims);
	}

	/** Used internally to translate coordinates into an index in the bitset. 
	 *  Provide your own dims version for adress computations.
	 * */
	private static int toAdr (int [] coords,int []dims) {
		assert (coords.length == dims.length);
		int n = 0;
		int total=1;
		for (int i=dims.length-1; i >= 0 ; i--) {
			n += total *coords[i];
			total *= dims[i];
		}
		return n;
	}

	/** Used internally to translate an index in the bitset into coordinates. 
	 *  See also toAdr() for reverse operation.
	 *  Uses this.dims for adress computations.
	 * */
	private int[] toCoord (int adr) {
		int [] coord = new int[dims.length];
		int total=1;
		for (int v : dims )
			total*= v;
		for (int i=0; i<dims.length ; i++) {
			total /= dims[i];
			coord [i] = adr / total ;
			adr %= total;			
		}
		return coord;
	}

	/** Creates a new tensor of rank increased by 1 : a new dimension 
	 * of size "dimsize" is inserted in position "dim" w.r.t. to the definition of the
	 * the current tensor dimensions. 
	 * The "this" tensor is inserted into this new tensor at position val of the newly added rank. 
	 */
	 public Tensor insert (int dim, int dimsize, int val) {
		int [] newdims = new int[dims.length+1];
		int j=0;
		for (int i=0; i<newdims.length ; i++) {
			if (i!= dim)
				newdims[i]=dims[j++];
			else
				newdims[i]=dimsize;
		}
		Tensor t = new Tensor(newdims);

		// handle null domain tensor == int
		if (dims.length == 0) {
			if (! vals.isEmpty())
				t.vals.set(val);
			return t;
		}
		// to hold new coords of current bit
		int [] coord = new int[newdims.length];
		// preliminary compute multiplicator for tocoord only once
		int dimsleft=1;
		for (int v : dims )
			dimsleft*= v;

		for(int adrit=vals.nextSetBit(0); adrit>=0; adrit=vals.nextSetBit(adrit+1)) {
			int adr=adrit;
			// copied from toCoord

			int tmptotal = dimsleft;
			// operate on index i here
			j = 0;
			for (int i=0; i<newdims.length ; i++) {
				if (i != dim){
					tmptotal /= dims[j];
					coord [i] = adr / tmptotal ;				
					// could maybe break+continue on i==dim and coord[i]!=i1 and i2
					adr %= tmptotal;
					j++;
				} else {
					coord[i] = val;
				}
			}
			t.add(coord);
		}		
		return t;

	}

	 /** Reverse of insert, extract the subtensor of rank : this.rank-1  such that
	  *  extract(1,dim) : A(i,j) = A(i,dim,j), extract(0,dim) :  A(i,j) = A(dim,i,j)...
	  *  Example, for a matrix extract(0,XX) returns line XX, extract(1,XX) returns column XX.
	  * returns the  */
	public Tensor extract (int rank, int dim) {
		assert(dims.length > 0);
		int [] newdims= new int[dims.length-1];
		int j=0;
		for (int i=0; i<dims.length ; i++) {
			if (i!= rank)
				newdims[j++]=dims[i];
		}
		Tensor t = new Tensor(newdims);
		//		to hold coords of current bit
		int [] coord = new int[dims.length];
		// preliminary compute multiplicator for tocoord only once
		int dimsleft=1;
		for (int v : dims )
			dimsleft*= v;

		for(int adrit=vals.nextSetBit(0); adrit>=0; adrit=vals.nextSetBit(adrit+1)) {
			int adr=adrit;
			// copied from toCoord

			int tmptotal = dimsleft;
			// operate on index i here
			for (int i=0; i<dims.length ; i++) {
				tmptotal /= dims[i];
				coord [i] = adr / tmptotal ;				
				// could maybe break+continue on i==dim and coord[i]!=i1 and i2
				adr %= tmptotal;			
			}
			if (coord[rank]==dim) {
				int []newcoord = new int[coord.length-1];
				j=0;
				for (int i=0; i<coord.length ; i++) {
					if (i!= rank)
						newcoord[j++]=coord[i];
				}
				t.add(newcoord);

			}
		}		
		return t;
	}
	/** Adds all the elements of the argument tensor to this.
	 * Tensors should be homogeneous (or you will get assert failure).
	 * */
	public void addAll(Tensor t) {
		assert (t.dims == this.dims);
		vals.or(t.vals);		
	}

	/** Returns true if all the elements of t are contained in this.
	 *  Tensors should be homogeneous (or you will get assert failure).
	 * */
	public boolean containsAll (Tensor t) {
		// or value based comparison
		assert (t.dims == this.dims);
		BitSet tmp = (BitSet) vals.clone();
		tmp.and(t.vals);
		return tmp.equals(t.vals);
	}

	/** Returns true if all the elements of t are contained in this.
	 *  Tensors should be homogeneous (or you will get assert failure).
	 * */
	public void removeAll(Tensor t) {
		// or value based comparison
		assert (t.dims == this.dims);
		vals.andNot(t.vals);
	}

	/** Something simple that displays the contents of the tensor. Used for debug.*/
	public String toString () {
		StringBuffer sb = new StringBuffer(255);
		for(int i=vals.nextSetBit(0); i>=0; i=vals.nextSetBit(i+1)) {
			// operate on index i here
			sb.append(Arrays.toString(toCoord(i))+",");
		}
		return sb.toString() +vals.toString() + Arrays.toString(dims);
	}

	/** Kill a dimension from a rank of the tensor.
	 * For instance in a matrix suppress(0,XXX) kills line XXX, suppress(1,XXX) kills column XXX. 
	 * @param rank the target dimension
	 * @param dim the value targeted along the dimension.
	 * @return a new tensor with updated dims
	 */
	public Tensor suppress (int rank, int dim) {
		int [] newdims = dims.clone();
		newdims[rank]--;

		Tensor t = new Tensor(newdims);
		// to hold coords of current bit
		int [] coord = new int[dims.length];
		// preliminary compute multiplicator for tocoord only once
		int dimsleft=1;
		for (int v : dims )
			dimsleft*= v;

		for(int adrit=vals.nextSetBit(0); adrit>=0; adrit=vals.nextSetBit(adrit+1)) {
			int adr=adrit;
			// copied from toCoord

			int tmptotal = dimsleft;
			// operate on index i here
			for (int i=0; i<dims.length ; i++) {
				tmptotal /= dims[i];
				coord [i] = adr / tmptotal ;
				// could maybe break+continue on i==dim and coord[i]!=i1 and i2
				adr %= tmptotal;			
			}
			if (coord[rank] < dim)
				t.add(coord);
			else if (coord[rank] > dim) {
				coord[rank]--;
				t.add(coord);				
			} else
				// no add
				continue;
		}


		return t;
	}

	/** compares two rows (faces in 3D) for equality.
	 * used in merge operation. 
	 * e.g. isEqual (C, 0, 1) tests if elems 0 and 1 of C can be merged.
	 * @param dim the number of the dimension concerned. For domain CxD, 0 designates C and 1 D.
	 * @param i1  first index
	 * @param i2  index to compare
	 * @return true if they are equal
	 */
	public boolean isEqual (int dim, int i1, int i2) {
		System.out.println(this);
		//		 compute into total the size of the parts to be compared
		int total=1;
		for (int k=0;k<dims.length; k++)
			if (k!=dim) total *= dims[k];

		if (dim==0) {
			// favorable case, no reordering is necessary


			int []coord = dims.clone();
			// extract i dimension
			coord[0]=i1;
			int starti = toAdr(coord);
			BitSet dimi = vals.get(starti, starti+total);
			// extract j dimension
			coord[0]=i2;
			int startj = toAdr(coord);
			BitSet dimj = vals.get(startj, startj+total+1);
			return dimi.equals(dimj);
		} else {
			// need to construct by restructuring the bitset
			BitSet dimi = new BitSet(total);
			BitSet dimj = new BitSet(total);
			// extracts of toCoord placed here to avoid multiple allocation of coord
			// to hold coords of current bit
			int [] coord = new int[dims.length];
			// preliminary compute multiplicator for tocoord only once
			int dimsleft=1;
			for (int v : dims )
				dimsleft*= v;

			// prepare structure for reconversion
			// note that swapdims[0] could be set to zero here: 
			// the dimension gets suppressed in the computation
			int [] swapdims = dims.clone();
			swapdims[dim] = dims[0];
			swapdims[0] = dims[dim];
			// to avoid a double test below
			BitSet toAdd;
			for(int adrit=vals.nextSetBit(0); adrit>=0; adrit=vals.nextSetBit(adrit+1)) {
				int adr=adrit;
				// copied from toCoord
				int tmptotal = dimsleft;
				// operate on index i here
				for (int i=0; i<dims.length ; i++) {
					tmptotal /= dims[i];
					coord [i] = adr / tmptotal ;
					// could maybe break+continue on i==dim and coord[i]!=i1 and i2
					adr %= tmptotal;			
				}

				// set toAdd
				if (coord[dim] == i1 )
					toAdd = dimi;
				else if ( coord[dim]== i2) 
					toAdd = dimj;
				else
					// skip this bit
					continue;

				//reach here only if this bit is concerned
				coord[dim] = coord[0];
				// set to 0 to destroy a dimension in the conversion
				coord[0]=0;
				// update appropriate tensor
				toAdd.set(toAdr(coord,swapdims));
			}
			return dimi.equals(dimj);

		}

	}


	/** Very basic test. TODO : get rid of this artefact */
	public static void main(String[] args) {
		int [] dims = {3,2,4};
		Tensor t = new Tensor(dims);
		System.out.println(t);
		int [] tok = {0,0,0};
		t.add(tok);
		System.out.println(t);
		int [] tok2 = {2,1,1};
		t.add(tok2);
		System.out.println(t);

		// test 1
		// 2x2 tensor, one entry = no matches


	}

	/** creates a new copy of this tensor. 
	 * dims is still shared with original, but bitset is cloned.
	 */ 
	public Tensor copy() {
		return new Tensor(dims,(BitSet) vals.clone());
	}

	/** eclipse generated hashCode. 
	 */
	public int hashCode() {
		final int PRIME = 31;
		int result = 1;
		result = PRIME * result + Arrays.hashCode(dims);
		result = PRIME * result + ((vals == null) ? 0 : vals.hashCode());
		return result;
	}

	/** eclipse generated equals.
	 * TODO : stronger assumptions on internal state of tensor allows less tests. Optimize. 
	 */
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final Tensor other = (Tensor) obj;
		if (!Arrays.equals(dims, other.dims))
			return false;
		if (vals == null) {
			if (other.vals != null)
				return false;
		} else 
			if (!vals.equals(((Tensor)obj).vals))
			return false;
		return true;
	}

	/** Returns the total number of (potential entries in the tensor.
	 * i.e. multiply the dimensions of the ranks.
	 */
	public int size() {
		int total =1;
		for (int i =0; i < dims.length; i++)
			total *= dims[i];
		return total;
	}

	/** Natural ordering for tensors. Based on a pure lexico order on bitset.
	 * Dimensions are NOT compared.
	 */
	public static int compareBitSet ( BitSet vals, BitSet other) {
		int adri = vals.nextSetBit(0);
		int adrj = other.nextSetBit(0);
	
		while (adri == adrj) {
			if (adri <0)
				return 0;
			adri=vals.nextSetBit(adri+1);
			adrj = other.nextSetBit(adrj+1);			
		}
		if (adri > adrj) 
			return 1;
		else return -1;
	}
	
	
	public int compareTo(Tensor other) {
		return compareBitSet(vals, other.vals);
	}

	/** returns true if the Tensor contains only 0 valued entries. */
	public boolean isEmpty() {
		return vals.isEmpty();
	}

	/** Allows to iterate on a tensor as if it was a set of coordinates. 
	 *  use in loop like 
	 *  ITensorIterator it = t.getIterator() 
	 *  while (ti.hasNext()) {
	 *       int [] coord = ti.next();
	 *		 //operate here
	 *	}
	 * */
	public ITensorIterator getIterator() {
		return new TensorIterator();
	}
	
	/// use in loop like 
	/// while (ti.hasNext()) {
	/// 	int [] coord = ti.next();
	/// 	// operate here
	/// }
	class TensorIterator implements ITensorIterator {
		int adrit = -1;
		int [] coord ;
		int dimsleft=1;
		
		public TensorIterator() {
			// to hold coords of current bit
			coord = new int[dims.length];
			// preliminary compute multiplicator for tocoord only once
			for (int v : dims )
				dimsleft*= v;
		}
		public boolean hasNext() {
			adrit=vals.nextSetBit(adrit+1);
			return adrit >=0;
		}

		public int [] next () {				 
			int adr = adrit;
			// copied from toCoord
			int tmptotal = dimsleft;
			// operate on index i here
			for (int i=0; i<dims.length ; i++) {
				tmptotal /= dims[i];
				coord [i] = adr / tmptotal ;
				adr %= tmptotal;			
			}
			return coord;
		}
		
		public void clear() {
			vals.clear(adrit);
		}

	}

	/** returns the number of dimensions along the axis rank. */
	public int getRankSize(int rank) {
		return dims[rank];
	}

//	/** Resize dims. 
//	 * TODO : lots of issues with this function, it is currently CHAOTIC 
//	 * EVIL and should not be used yet. 
//	 * You will get assertion failures + it modifies dims when every other
//	 *  function rely on the fact that it is constant. */
//	public void ensureRankSize(int i, int index) {
//		if (dims.length < i) {
//			System.err.println("resize of dims not implmented");
//			assert(false);
//		}
//		if (i!=0) {
//			System.err.println("resize of dims not implmented for rank other than 0");
//			assert(false);			
//		}
//		// simplest case, adds 0 to end of bitset = noop
//		if (dims[i] < index)
//			dims[i]=index;
//		// the bitset is fine, it resizes by itself.
//		
//	}

	public void setAll() {
		vals.set(0, size());
	}


}
